// -*- c++ -*-
/*
 *  MICO --- an Open Source CORBA implementation
 *  Copyright (c) 1997-2007 by The Mico Team
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Library General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Library General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  For more information, visit the MICO Home Page at
 *  http://www.mico.org/
 */

#ifndef __mico_dii_h__
#define __mico_dii_h__

namespace CORBA {

// environment
class MICO_EXPORT Environment : public ServerlessObject {
    Exception *_exceptn;
    Principal_ptr _princ;
public:
    void exception (Exception *);
    Exception *exception () const
    { return _exceptn; }

    void clear ();

    // begin-mico-extension
    Environment (Principal_ptr = NULL);
    ~Environment ();

    Principal_ptr principal ()
    { return _princ; }
    // end-mico-extension

    static Environment_ptr _duplicate (Environment_ptr o)
    {
	if (o)
	    o->_ref();
	return o;
    }
    static Environment_ptr _nil ()
    {
	return 0;
    }
};

typedef ObjVar<Environment> Environment_var;


// named value
class NamedValue : public ServerlessObject {
    String_var _n;
    Any _v;
    Flags _f;
public:
    const char *name () const
    { return _n.in(); }

    Any *value () const
    { return (Any *)&_v; }

    Flags flags () const
    { return _f; }

    // begin-mico-extension
    MICO_EXPORT NamedValue (const char *name, const Any &, Flags f = 0);
    MICO_EXPORT NamedValue (const char *name, Flags f = 0);
    MICO_EXPORT NamedValue (); // vector<NamedValue> needs this
    MICO_EXPORT NamedValue &operator= (const NamedValue &);
    MICO_EXPORT ~NamedValue ();
    // end-mico-extension

    static NamedValue_ptr _duplicate (NamedValue_ptr o)
    {
	if (o)
	    o->_ref();
	return o;
    }
    static NamedValue_ptr _nil ()
    {
	return 0;
    }
};

typedef ObjVar<NamedValue> NamedValue_var;
typedef ObjOut<NamedValue> NamedValue_out;

enum NVListFlags {
    IN_COPY_VALUE = 0x0001,
    DEPENDENT_LIST = 0x0002
};

// nv list
class NVList : public ServerlessObject {
    std::vector<NamedValue_var> _vec;
public:
    typedef CORBA::Bounds Bounds;
    typedef CORBA::Bounds_catch Bounds_catch;

    ULong count () const
    { MICO_OBJ_CHECK (this); return _vec.size(); }

    MICO_EXPORT NamedValue_ptr add (Flags);
    MICO_EXPORT NamedValue_ptr add_item (const char *, Flags);
    MICO_EXPORT NamedValue_ptr add_value (const char *, const Any &, Flags);
    MICO_EXPORT NamedValue_ptr add_item_consume (char *, Flags);
    MICO_EXPORT NamedValue_ptr add_value_consume (char *, Any *, Flags);
    MICO_EXPORT NamedValue_ptr item (ULong idx);
    MICO_EXPORT void remove (ULong);

    // begin-mico-extension
    MICO_EXPORT NVList (ULong count = 0);
    MICO_EXPORT ~NVList ();

    MICO_EXPORT CORBA::Boolean copy (NVList_ptr, Flags flags);
    // end-mico-extension

    static NVList_ptr _duplicate (NVList_ptr o)
    {
	if (o)
	    o->_ref();
	return o;
    }
    static NVList_ptr _nil ()
    {
	return 0;
    }
};

typedef ObjVar<NVList> NVList_var;
typedef ObjOut<NVList> NVList_out;


// exception list
class ExceptionList : public ServerlessObject {
    std::vector<TypeCode_var> _vec;

public:
    typedef CORBA::Bounds Bounds;
    typedef CORBA::Bounds_catch Bounds_catch;

    ULong count () const
    { MICO_OBJ_CHECK (this); return _vec.size(); }

    MICO_EXPORT void add (TypeCode_ptr tc);
    MICO_EXPORT void add_consume (TypeCode_ptr tc);
    MICO_EXPORT TypeCode_ptr item (ULong idx);
    MICO_EXPORT void remove (ULong idx);

    // begin-mico-extension
    MICO_EXPORT ExceptionList ();
    MICO_EXPORT ~ExceptionList ();
    // end-mico-extension

    static ExceptionList_ptr _duplicate (ExceptionList_ptr o)
    {
	if (o)
	    o->_ref();
	return o;
    }
    static ExceptionList_ptr _nil ()
    {
	return 0;
    }
};

typedef ObjVar<ExceptionList> ExceptionList_var;
typedef ObjOut<ExceptionList> ExceptionList_out;


// context list
class ContextList : public ServerlessObject {
    std::vector<std::string> _vec;

public:
    typedef CORBA::Bounds Bounds;
    typedef CORBA::Bounds_catch Bounds_catch;

    ULong count () const
    { MICO_OBJ_CHECK (this); return _vec.size(); }

    MICO_EXPORT void add (const char *ctxt);
    MICO_EXPORT void add_consume (char *ctxt);
    MICO_EXPORT const char *item (ULong idx);
    MICO_EXPORT void remove (ULong idx);

    // begin-mico-extension
    MICO_EXPORT ContextList ();
    MICO_EXPORT ~ContextList ();
    // end-mico-extension

    static ContextList_ptr _duplicate (ContextList_ptr o)
    {
	if (o)
	    o->_ref();
	return o;
    }
    static ContextList_ptr _nil ()
    {
	return 0;
    }
};

typedef ObjVar<ContextList> ContextList_var;


enum ArgModeFlags {
    ARG_IN = 0x0001,
    ARG_OUT = 0x0002,
    ARG_INOUT = 0x0004
};

enum RequestFlags {
    OUT_LIST_MEMORY = 0x0001,
    INV_NO_RESPONSE = 0x0002,
    INV_TERM_ON_ERR = 0x0004,
    RESP_NO_WAIT = 0x0008
};

struct RequestCallback {
    enum Event { RequestDone };
    virtual void callback (Request_ptr, Event) = 0;
    virtual ~RequestCallback ();
};

// request
class Request : public ServerlessObject, public ORBCallback {

    Object_var _object, _original_object;
    bool _is_forwarded;
    Context_ptr _context;
    String_var _opname;
    NVList_ptr _args;
    NamedValue_ptr _res;
    Flags _flags;
    ExceptionList_ptr _elist;
    ContextList_ptr _clist;
    Environment_ptr _environm;
    CORBA::ORBMsgId _orbid;
    Boolean _invoke_pending;
    ORBRequest *_orbreq;
    RequestCallback *_cb;
    // PI client interceptor request info
    PInterceptor::ClientRequestInfo_impl* _cri;
    CORBA::Boolean _invoke_flag;
#ifdef HAVE_THREADS
    ORBAsyncCallback async_callback;
#endif

    Request (const Request &r)
        : CORBA::ServerlessObject(r),
          CORBA::ORBCallback()
    { assert (0); }

    Request &operator= (const Request &) { assert (0); return *_nil(); }

    MICO_EXPORT void decode_user_exception ();

    MICO_EXPORT virtual Boolean waitfor (ORB_ptr orb, CORBA::ORBMsgId id, ORBCallback::Event ev, Long tmout = -1);
    MICO_EXPORT virtual void notify (ORB_ptr orb, CORBA::ORBMsgId id, ORBCallback::Event ev);
public:
    void _forward (Object_var obj)
    {
      _is_forwarded = true;
      _object = obj;
    }
    void _unforward()
    {
      _is_forwarded = false;
      _object = _original_object;
    }

    Object_ptr target () const
    { return _object.in(); }

    const char *operation () const
    { return _opname; }

    NVList_ptr arguments ()
    { MICO_OBJ_CHECK (this); return _args; }

    NamedValue_ptr result ()
    { return _res; }

    Environment_ptr env ()
    { return _environm; }

    ExceptionList_ptr exceptions ()
    { return _elist; }

    ContextList_ptr contexts ()
    { return _clist; }
    
    void ctx (Context_ptr);
    Context_ptr ctx () const
    { return _context; }

    MICO_EXPORT void invoke ();
    MICO_EXPORT void send_oneway ();
    MICO_EXPORT void send_deferred (RequestCallback *cb = 0);
    MICO_EXPORT void get_response (Boolean block = TRUE);
    MICO_EXPORT Boolean poll_response ();

    MICO_EXPORT Any &add_in_arg ();
    MICO_EXPORT Any &add_in_arg (const char *name);
    MICO_EXPORT Any &add_inout_arg ();
    MICO_EXPORT Any &add_inout_arg (const char *name);
    MICO_EXPORT Any &add_out_arg ();
    MICO_EXPORT Any &add_out_arg (const char *name);
    MICO_EXPORT void set_return_type (TypeCode_ptr);
    MICO_EXPORT Any &return_value ();

    // begin-mico-extension
    MICO_EXPORT Request (Object_ptr, Context_ptr, const char *op,
	     NVList_ptr args, NamedValue_ptr result,
	     Flags f);
    MICO_EXPORT Request (Object_ptr, Context_ptr, const char *op,
	     NVList_ptr args, NamedValue_ptr result,
	     ExceptionList_ptr, ContextList_ptr,
	     Flags f);
    MICO_EXPORT Request (Object_ptr, const char *op);
    MICO_EXPORT Request (); // vector<Request> needs this
    MICO_EXPORT ~Request ();
    // end-mico-extension

    static Request_ptr _duplicate (Request_ptr o)
    {
	if (o)
	    o->_ref();
	return o;
    }
    static Request_ptr _nil ()
    {
	return 0;
    }
};

typedef ObjVar<Request> Request_var;


enum ContextFlags {
    CTX_DELETE_DESCENDENTS = 0x0001,
    CTX_RESTRICT_SCOPE = 0x0002
};

typedef ObjVar<Context> Context_var;
typedef ObjOut<Context> Context_out;

// context
class Context : public ServerlessObject {
    Context_ptr _myparent;
    std::vector<Context_var> _childs;
    NVList_var _properties;
    std::string _name;

    MICO_EXPORT Boolean match (const char *value, const char *pattern);
public:
    MICO_EXPORT const char *context_name () const;
    MICO_EXPORT Context_ptr parent () const;

    MICO_EXPORT void create_child (const char *, Context_out);

    MICO_EXPORT void set_one_value (const char *, const Any &);
    MICO_EXPORT void set_values (NVList_ptr);
    MICO_EXPORT void delete_values (const char *);
    MICO_EXPORT void get_values (const char *, Flags, const char *, NVList_out);
    MICO_EXPORT void get_values (const char *, Flags, ContextList_ptr, NVList_out);

    // begin-mico-extension
    MICO_EXPORT Context (DataDecoder &);
    MICO_EXPORT Context (const char *name, Context_ptr parent = 0);
    MICO_EXPORT ~Context ();

    MICO_EXPORT void encode (DataEncoder &, ContextList_ptr = 0) const;
    MICO_EXPORT Boolean decode (DataDecoder &);
    // end-mico-extension

    static Context_ptr _duplicate (Context_ptr o)
    {
	if (o)
	    o->_ref();
	return o;
    }
    static Context_ptr _nil ()
    {
	return 0;
    }
};


// principal
class Principal : public ServerlessObject {
    std::vector<Octet> _rep;
    static std::vector<Octet> _peer_info;
    CORBA::Transport *_transp;
public:
    // begin-mico-extension
    MICO_EXPORT Principal (DataDecoder &, CORBA::Transport * = 0);
    MICO_EXPORT Principal (CORBA::Transport * = 0);
    MICO_EXPORT virtual ~Principal ();
#ifdef USE_CSL2
    MICO_EXPORT CORBA::Transport* transport(); // ###ras
#endif /* USE_CSL2  */
    MICO_EXPORT CORBA::Boolean operator== (const Principal &) const;

    MICO_EXPORT void encode (DataEncoder &) const;
    MICO_EXPORT Boolean decode (DataDecoder &);

    typedef StringSequenceTmpl<String_var> PropertyNameList;
    typedef PropertyNameList *PropertyNameList_ptr;
    typedef TVarVar<PropertyNameList> PropertyNameList_var;
    typedef TVarOut<PropertyNameList> PropertyNameList_out;

    MICO_EXPORT virtual PropertyNameList_ptr list_properties ();
    MICO_EXPORT virtual Any* get_property (const char *prop_name);
#ifdef USE_CSL2
    MICO_EXPORT virtual Any* get_attribute(unsigned long fam, unsigned long attr); // ###ras
    MICO_EXPORT virtual CORBA::Boolean check(std::string& s);
#endif /* USE_CSL2  */

    MICO_EXPORT static void set_peer_info (const OctetSeq &);
    // end-mico-extension

    static Principal_ptr _duplicate (Principal_ptr o)
    {
	if (o)
	    o->_ref();
	return o;
    }
    static Principal_ptr _nil ()
    {
	return 0;
    }
};

typedef ObjVar<Principal> Principal_var;

}

#endif // __mico_dii_h__
